#module NNTP_TCPUCXM "V1.1"

/* Multithreaded NNTP server for ANU_NEWS V6.x and VMS UCX 1.3A
 * Written by Steve Bour (jsbour@ualr.edu).
 * taken from the NNTP server for CMU-TEK TCP/IP V6.5 by Keith Halewood
 * The 'getremhost' routine has been stolen from the original single-thread
 * server interface.
 *
 * Changes made CMU-TEK version -> VMS UCX version (by Steve Bour)
 * 1. Changed QIOs from CMU-TEK to UCX (minimal hacking required, some
 *    required no changes at all).
 * 2. Added in NNTP inactivity timeout functionality missing from UCX
 *    by use of an "inactivity timer tick", forcing a SYS$CANCEL on any
 *    thread that is inactive (no reads or writes) for INACTIVITY_TICK_MAX
 *    ticks.  Currently the ticks are 900 seconds (15 minutes), and the
 *    SYS$CANCEL is executed on any channel that's inactive for 10 ticks
 *    (~ 2.5 hours).  It would have been better, perhaps, to have seperate
 *    AST timers on each active channel, but my VMS system call experience
 *    is weak (read: too lazy to figure it out :-).  The UCX driver doesn't
 *    appear to support the CMU-TEK style timeout parameter in the QIOs, thus
 *    this kludge.
 * 3. Added code to log low-level driver activity to a file; some of this is
 *    commented out and when I'm feeling really good about *my* part of the
 *    code (if that's possible :-) it'll probably all be ripped out.
 * 4. I didn't implement RFC931 authentication -- if anyone needs it,
 *    harass me and I might put it in.
 * 5. DCL command to start up server (executed from the News Manager account):
 *
 *               $ run/detach -
 *                     /output = NLA0: -
 *                     /error = NLA0: -
 *                     /input = NLA0: -
 *                     /process_name = "UCX/NNTP server" -
 *                     NEWS_MANAGER_DEV:[NEWS_DIST]NNTP_TCPUCXM.EXE
 *
 * 6. Many thanks to Keith Halewood for the base code.
 *    
 * Changes V1.0 -> V1.1 (CMU-TEK version, by Keith Halewood)
 * 1. Server exits when the passive listen times out and there are no
 *    active connections. Previously, the server exited immediately on
 *    the termination of the last active connection.
 * 2. As a consequence of the above, a bug preventing threads from
 *    being freed correctly has now been fixed.
 * 3. Rearranged manage_open so that a new passive listen is queued
 *    before the new thread is initialised to minimise the chances of
 *    of missing closely-spaced incoming connect requests.
 * 4. getremhost() has been modified to perform RFC931 authentication
 *    checking. Provided that a daemon is running on the client host,
 *    NNTP can now log the user as well as the host (a la DECnet).
 *    It's the result of 10 mins work so the code isn't pretty.
 * 5. Modified write_thread so that multiple lines are blocked into one
 *    QIO operation saving lots of CPU and increasing I/O throughput.
 * 6. log_to_file() is now called (see nntp_server source code).
 * 7. Added call to already_running which uses the channel count on
 *    a permanent mailbox to determine whether a nntp_server is already
 *    running. It is currently programmed to terminate if one is
 *    already active.
 * 8. Fixed read buffer overflow into read residue for each thread. -1
 *    should be +1 in array declarations.
 * 8. Manage_read_completion more efficient.
 */

#include <iodef.h>
#include <dvidef.h>
#include <signal.h>
#include <stdio.h>
#include <descrip.h>
#include <errno.h>
#include <types.h>
#include <socket.h>
#include <in.h>
#include <netdb.h>
#include <inet.h>
#include <ucx$inetdef.h>
#include <time.h>
  
#define PORT        119    /* Should be 119, use 8888 for testing. */
#define MAXTHREADS   16    /* Maximum number of services - Make sure
                            * process quotas are capable of handling 
                            * the number chosen.
                            */
#define AUTHTIMEOUT  30    /* Authentication timeout.
                            * Some Un*x auth servers are very slow, but
                            * none should be slower than this.
                            */
#define BUFSIZE    1024    /* Constants for QIO buffer lengths and
                            */
#define LINESIZE   2048    /* the read line assembly buffer.
                            */

/* Event numbers used by AST completion routines. The are placed,
 * along with a thread number into the attention queue.
 */

#define EV_NULL                     0
#define EV_OPEN_COMPLETED           1
#define EV_OPEN_ABORTED             2
#define EV_READ_COMPLETED           3
#define EV_READ_ABORTED             4
#define EV_WRITE_COMPLETED          5
#define EV_WRITE_ABORTED            6

/* Is the thread in use (healthy) or unassigned. Unassigned is also
 * a state used during shutdown to control the discarding of entries
 * in the I/O queues.
 */

#define ACT_UNASSIGNED              0
#define ACT_HEALTHY                 1

/* 'Next function' return codes, determines whether the 'next function' call
 * for a thread is activated on receipt of a line (command) or on receipt of
 * the file terminator (a single full-stop) or whether the thread is to be
 * closed down.
 */

#define TYP_NO_INPUT                0
#define TYP_CMD_INPUT               1
#define TYP_FILE_INPUT              2

/* Definitions for the passive open (listen) on PORT
 */

unsigned short listen_channel, clone_channel, listen_iosb[4], new_iosb[4];
int active_count = 0;
int active_listener = 0;

/* I/O queue definition.
 */

struct line_block
  {
    struct line_block *next;
    char *line;
    int len;
  };

/* Thread structures. [0] isn't used.
 */

volatile static struct thread_struct
  {
    int active;
    unsigned short channel;
    unsigned short write_iosb[4], read_iosb[4];
    char write_buffer[BUFSIZE+1], read_buffer[BUFSIZE+1];
    char read_residue[LINESIZE+1];
    struct line_block *read_head, *read_tail, *write_head, *write_tail;
    int write_in_progress;
    int next_type;
    int (*next_function)();
    char remhost[255];
    char remuser[32];
    int inactivity_ticks;
  } threads[MAXTHREADS+1];

struct dsc$descriptor ipdsc = { 3, DSC$K_CLASS_S, DSC$K_DTYPE_T, "BG:"};
/* I couldn't seem to get UCX$DEVICE to work; not only that, but BG: is
   what *DIGITAL* uses in their example code.  Hmmmm.............. jsb */

/* NNTP inactivity timeout definitions
 */

#define INACTIVITY_TICK_LENGTH 900	  /* how many seconds between ticks */
#define INACTIVITY_TICK_MAX 10		  /* how many ticks before abort */
					  /* 900 * 10 = 9000, or 2.5 hours */
					  /* (minus up to 15 minutes) */
static int inactivity_tick_trap_flag = 0; /* gets set to 1 when tick occurs */

/* code for logging server driver events
 */

FILE *drvrlog;
#define drvrlogfile "NEWS_MANAGER_DEV:[NEWS_LOG]NNTP_SERVER_DRVR.LOG"

/* Definition for time of day routines
 */

time_t currtime;

/* stuff for listener initialization
 */

struct mysockaddr 
  {
    short inet_family;
    short inet_port;
    unsigned char adrs[4];
    char blkb[8];
  };

struct mysockaddr2
  {
    short   inet_family;
    short   inet_port;
    int     adrs;
    char    blkb[8];
  };

struct itlst
  {
    int lgth ;
    struct mysockaddr2 *hst ;
  };

struct itlst_3
  {
    int lgth ;
    struct mysockaddr *hst ;
    int *retlth;
  };

struct mysockaddr remote_hostaddr;

int retlen;
int beenhere=0;
struct mysockaddr2 local_host;
struct itlst lhst_adrs;
struct itlst_3 rhst_adrs;
short sck_parm[2] = { INET$C_TCP, INET_PROTYP$C_STREAM };

/******************************************************************
 * Exteral routines from NNTP_SERVER                              *
 ******************************************************************/

int server_init();
void server_init_unit();
void log_to_file();

/******************************************************************
 * Thread event queue routines - ast interlocked (ie. safe)       *
 ******************************************************************/

#define MAXQUEUE 255                   /* Arbitrary but large */

volatile static int queue_thread_array[MAXQUEUE];
volatile static int queue_event_array[MAXQUEUE];
volatile static int queue_count = 0;
volatile static int queue_head = 0;
volatile static int queue_tail = 0;

static int enqueue(thread,event)
int thread, event;
{
  int result = 1;
  (void)sys$setast(0);
  if (queue_count == MAXQUEUE)
  {
    result = 0;
  }
  else
  {
    queue_count++;
    queue_thread_array[queue_head] = thread;
    queue_event_array[queue_head++] = event;
    if (queue_head==MAXQUEUE) queue_head = 0;
  }
  (void)sys$setast(1);
  return(result);
}

static int dequeue(thread,event)
int *thread, *event;
{
  int result = 1;
  (void)sys$setast(0);
  if (queue_count > 0)
  {
    *thread = queue_thread_array[queue_tail];
    *event = queue_event_array[queue_tail++];
    queue_count--;
    if (queue_tail==MAXQUEUE) queue_tail = 0;
  } else
  {
    result = 0;
  }
  (void)sys$setast(1);
  return(result);
}

/* Convenient way of duplicating a string
 */

char *strdup(s)
char *s;
{
  char *t = (char *)malloc(sizeof(char)*(strlen(s)+1));
  strcpy(t,s);
  return(t);
}

/***************************************************************
 * AST handlers                                                *
 ***************************************************************/

void ast_signal_open_complete(dummy)
int dummy;
{
  (void)enqueue(0,(listen_iosb[0]&1) ? EV_OPEN_COMPLETED : EV_OPEN_ABORTED);
  (void)sys$wake(0,0);
}

void ast_signal_read_complete(thread)
int thread;
{
  if (threads[thread].active==ACT_HEALTHY)
  {
    (void)enqueue(thread, (threads[thread].read_iosb[0]&1) ?
                          EV_READ_COMPLETED : EV_READ_ABORTED);
    (void)sys$wake(0,0);
  }
}

void ast_signal_write_complete(thread)
int thread;
{
  if (threads[thread].active==ACT_HEALTHY)
  {
    (void)enqueue(thread, (threads[thread].write_iosb[0]&1) ?
                          EV_WRITE_COMPLETED : EV_WRITE_ABORTED );
    (void)sys$wake(0,0);
  }
}

inactivity_tick_trap()
{
  inactivity_tick_trap_flag = 1;
}

/***************************************************************
 * Read/Write queueing routines                                *
 ***************************************************************/

static void read_thread(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  int status;
  /* IO$_READVBLK reads a block of data (upto BUFSIZE) from an
   * open channel as in I/O users guide.
   */
  status = sys$qio(0,t->channel,IO$_READVBLK,t->read_iosb,
                   ast_signal_read_complete,thread,
                   t->read_buffer,BUFSIZE,0,0,0,0);
  if (!(status&1)) (void)enqueue(thread,EV_READ_ABORTED);
}

static void write_thread(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  struct line_block *b = t->write_head;
  int status, iosize, s;
  if ((t->active==ACT_HEALTHY) && (t->write_in_progress==0))
  {
    if (b!=NULL)
    {
      iosize = 0;
      t->write_buffer[0] = '\0';
      while (b)
      {
        s = strlen(b->line);
        if (s+iosize > BUFSIZE-2) break;
        strcat(t->write_buffer,b->line);
        iosize = iosize + s;
        t->write_head = b->next;
        free(b->line);
        free(b);
        b = t->write_head;
      }
      /* IO$_WRITEVBLK writes a block of data (length in P2) to an
       * open channel as in I/O users guide.
       */

      status = sys$qio(0,t->channel,IO$_WRITEVBLK,t->write_iosb,
                       ast_signal_write_complete,thread,
                       t->write_buffer,iosize,0,0,0,0);
      t->write_in_progress = status&1;
      if (!(status&1)) (void)enqueue(thread,EV_WRITE_ABORTED);
    }
  }
}

/***********************************************************************
 * Inits                                                               *
 ***********************************************************************/

static void init_thread(thread,channel)
int thread;
unsigned short channel;
{
  struct thread_struct *t = &threads[thread];
  t->active = ACT_HEALTHY;
  t->channel = channel;
  t->read_head = NULL; t->read_tail = NULL;
  t->write_head = NULL; t->write_tail = NULL;
  t->write_in_progress = 0;
  t->read_residue[0] = '\0';
  t->inactivity_ticks = 0;
  server_init_unit(thread);
  read_thread(thread);
  write_thread(thread);
}

static int find_clear_thread()
{
  int i;
  for (i=1;i<=MAXTHREADS;i++)
    if (threads[i].active==ACT_UNASSIGNED) return(i);
  return(0);
}

static void init_listener()
{
  int status;

  active_listener = 0;
  if (!beenhere)
  {
    lhst_adrs.lgth = sizeof local_host;
    lhst_adrs.hst = &local_host;
    local_host.inet_port = htons(PORT);
    local_host.inet_family = INET$C_AF_INET ;/* INET family */
    local_host.adrs = INET$C_INADDR_ANY; /* Network/subnetwork*/
    local_host.blkb[0] = 0 ;
    local_host.blkb[1] = 0 ;
    local_host.blkb[2] = 0 ;
    local_host.blkb[3] = 0 ;
    local_host.blkb[4] = 0 ;
    local_host.blkb[5] = 0 ;
    local_host.blkb[6] = 0 ;
    local_host.blkb[7] = 0 ;

    rhst_adrs.lgth = 16;
    rhst_adrs.hst = &remote_hostaddr;
    rhst_adrs.retlth = &retlen;
    retlen=0;

    /* set up listener channel */
    status = sys$assign(&ipdsc,&listen_channel,0,0);
    if (!(status&1))
    {
      drvrlog = fopen(drvrlogfile,"a+");
      fprintf(drvrlog,"Failed to assign Listener channel\n");
      fclose(drvrlog);
      return;
    }
    status = sys$qiow(0,listen_channel,IO$_SETMODE,listen_iosb,
                 0,0,&sck_parm,0,&lhst_adrs,3,0,0);
    if (!(status&1))
    {
      drvrlog = fopen(drvrlogfile,"a+");
      fprintf(drvrlog,"Failed to bind Listener channel\n");
      fclose(drvrlog);
      return;
    }
    beenhere=1;
  }
  status = sys$assign(&ipdsc,&clone_channel,0,0);
  if (!(status&1))
  {
    drvrlog = fopen(drvrlogfile,"a+");
    fprintf(drvrlog,"Failed to assign Clone channel\n");
    fclose(drvrlog);
    (void)enqueue(0,EV_OPEN_ABORTED);
  }
  status = sys$qio(0,listen_channel,IO$_ACCESS|IO$M_ACCEPT,listen_iosb,
                 ast_signal_open_complete,0,0,0,&rhst_adrs,&clone_channel,0,0);
  if (!(status&1))
  {
    drvrlog = fopen(drvrlogfile,"a+");
    fprintf(drvrlog,"Failed to bind Clone channel\n");
    fclose(drvrlog);
    (void)enqueue(0,EV_OPEN_ABORTED);
  }
  else
  {
    active_listener = 1;
  }
}

static void init_threads()
{
  struct thread_struct *t;
  int c;
  active_count = 0;
  for (t=threads,c=0;c<MAXTHREADS;t++,c++) t->active = ACT_UNASSIGNED;
}

init_inactivity_tick()
{
  extern inactivity_tick_trap();

  signal(SIGALRM,inactivity_tick_trap);
  alarm(INACTIVITY_TICK_LENGTH);
}

static int already_running()
{
struct dsc$descriptor mbxdesc = 
           { 3, DSC$K_CLASS_S, DSC$K_DTYPE_T, "NNTP_ANCHOR"};
  static unsigned short mbxchan;
  int status;
  int chncnt = 0;
  struct il
  { unsigned short buflen, itemcode;
    int *bufadr, *retlenadr; } itmlst[2];
  status = sys$crembx(1,&mbxchan,0,0,0xff,0,&mbxdesc);
  if (!(status&1)) return(0);
  (void)sys$delmbx(mbxchan);
  itmlst[0].buflen = 4;
  itmlst[0].itemcode = DVI$_REFCNT;
  itmlst[0].bufadr = &chncnt;
  itmlst[0].retlenadr = 0;
  itmlst[1].buflen = 0; itmlst[1].itemcode = 0;
  status = sys$getdviw(0,mbxchan,0,itmlst,0,0,0,0);
  if (!(status&1)) return(0);
  return(chncnt - 1);
}

static void init()
{
  if (already_running()) exit(1);
  if (!server_init(MAXTHREADS)) exit(1);
  init_threads();
  init_listener();
  init_inactivity_tick();
}

/***********************************************************************
 * Active routines                                                     *
 ***********************************************************************/

static void enqueue_reader(thread,string,len)
int thread;
char *string;
int len;
{
  struct thread_struct *t = &threads[thread];
  struct line_block *b;
  if (t->active==ACT_HEALTHY)
  {
    b = (struct line_block *) malloc(sizeof(struct line_block));
    b->next = NULL;
    b->line = strdup(string);
    b->len = len;
    if (t->read_head!=NULL)
    {
      t->read_tail->next = b;
      t->read_tail = b;
    }
    else
    {
      t->read_tail = b;
      t->read_head = b;
    }
    if ((t->next_type==TYP_CMD_INPUT) ||
        ((t->next_type==TYP_FILE_INPUT) && !strcmp(string,".")))
    {
      (t->next_function)(thread);
      write_thread(thread);
    }
  }
}

static void abort_thread(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  struct line_block *b;
  if (t->active != ACT_UNASSIGNED)
  {
    t->active = ACT_UNASSIGNED;
    b = t->read_head;
    while(b!=NULL)
    {
      struct line_block *c = b;
      b=b->next;
      free(c->line);
      free(c);
    }
    t->read_head = NULL; t->read_tail = NULL;
    b = t->write_head;
    while(b!=NULL)
    {
      struct line_block *c = b;
      b=b->next;
      free(c->line);
      free(c);
    }
    t->write_head = NULL; t->write_tail = NULL;
    if (t->next_type==TYP_FILE_INPUT) enqueue_reader(thread,".",1);
    if (t->next_type==TYP_CMD_INPUT) enqueue_reader(thread,"QUIT",4);
    log_to_file(thread);
    (void)sys$dassgn(t->channel);
    active_count--;
    currtime = time(NULL);
    drvrlog = fopen(drvrlogfile,"a+");
    fprintf(drvrlog,"%.24s: Thread %d aborted, %d active\n",
            asctime(localtime(&currtime)),
            thread,
            active_count);
    fclose(drvrlog);
    /* if (active_count<=0) exit(1); */
  }
}

static void close_thread(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  /* deaccess channel, don't flush pending i/o
   */
  t->active = ACT_UNASSIGNED;
  log_to_file(thread);
  (void)sys$qiow(0,t->channel,IO$_DEACCESS,t->write_iosb,0,0,0,0,0,0,0,0);
  (void)sys$dassgn(t->channel);
  active_count--;
  currtime = time(NULL);
  drvrlog = fopen(drvrlogfile,"a+");
  fprintf(drvrlog,"%.24s: Thread %d closed, %d active\n",
            asctime(localtime(&currtime)),
            thread,
            active_count);
  fclose(drvrlog);
/* if (active_count<=0) exit(1); */
}

static void manage_open()
{
  int thread, status;
  unsigned short new_channel;
  char resource_msg[] = "400 Server resource limits exceeded.\r\n";

  active_listener = 0;
  new_channel = clone_channel;
  init_listener();
  thread = find_clear_thread();
  if (thread>0)
  {
    init_thread(thread,new_channel);
    active_count++;
    currtime = time(NULL);
    drvrlog = fopen(drvrlogfile,"a+");
    fprintf(drvrlog,"%.24s: Thread %d opened, %d.%d.%d.%d %d, %d active\n",
            asctime(localtime(&currtime)),
            thread,
            remote_hostaddr.adrs[0],
            remote_hostaddr.adrs[1],
            remote_hostaddr.adrs[2],
            remote_hostaddr.adrs[3],
            ntohs(remote_hostaddr.inet_port),
            active_count);
    fclose(drvrlog);
}
  else
  {
    status = sys$qiow(0,new_channel,IO$_WRITEVBLK,new_iosb,0,0,
                      resource_msg,strlen(resource_msg),0,0,0,0);
    status = sys$qiow(0,new_channel,IO$_DEACCESS,listen_iosb,0,0,0,0,0,0,0,0);
    (void)sys$dassgn(new_channel);
  }
}

static void manage_failed_open()
{
  drvrlog = fopen(drvrlogfile,"a+");
  currtime = time(NULL);
  fprintf(drvrlog,"%.24s: manage_failed_open, %d total active threads\n",
                   asctime(localtime(&currtime)),
                   active_count);
  fclose(drvrlog);
  (void)sys$dassgn(clone_channel);
  active_listener = 0;
  if (active_count<=0) exit(1);
  init_listener();
}

static void manage_read_completion(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  char *cp, *cs;
  t->inactivity_ticks = 0; /* reset inactivity tick counter */
  t->read_buffer[t->read_iosb[1]]='\0';
  strcat(t->read_residue,t->read_buffer);
  cs = t->read_residue;
  while ((cp = strchr(cs,'\n'))!=NULL)
  {
    int len=cp-cs;

    *cp='\0';
    if (cp[-1]=='\r') {
	cp[-1]='\0';
	len--;
    }
    cp++;
    if (t->next_type!=TYP_NO_INPUT) enqueue_reader(thread,cs,len);
    cs = cp;
  }
  if (cs!=t->read_residue) strcpy(t->read_residue,cs);
  if (t->next_type!=TYP_NO_INPUT) read_thread(thread);
}

static void manage_failed_read(thread)
int thread;
{
  abort_thread(thread);
}

static void manage_failed_write(thread)
int thread;
{
  abort_thread(thread);
}

static void manage_write_completion(thread)
int thread;
{
  struct thread_struct *t = &threads[thread];
  t->inactivity_ticks = 0; /* reset inactivity tick counter */
  t->write_in_progress = 0;
  if (t->write_head)
  {
    write_thread(thread);
  }
  else
  {
    if (t->next_type==TYP_NO_INPUT) close_thread(thread);
  }
}

static void drain_queue()
{
  int thread, event;
  while (dequeue(&thread,&event))
  {
    switch (event)
    {
       case EV_OPEN_COMPLETED:  manage_open(); break;
       case EV_OPEN_ABORTED:    manage_failed_open(); break;
       case EV_READ_COMPLETED:  manage_read_completion(thread); break;
       case EV_READ_ABORTED:    manage_failed_read(thread); break;
       case EV_WRITE_COMPLETED: manage_write_completion(thread); break;
       case EV_WRITE_ABORTED:   manage_failed_write(thread); break;
    }
  }
}

check_for_inactive_threads()
{
  extern inactivity_tick_trap();
  struct thread_struct *t;
  int i;

  inactivity_tick_trap_flag=0;
  signal(SIGALRM,inactivity_tick_trap);
  alarm(INACTIVITY_TICK_LENGTH);

  for(i=1;i<=MAXTHREADS;i++)
  {
    t = &threads[i];
    if (t->active == ACT_HEALTHY)
    {
      if(++(t->inactivity_ticks) > INACTIVITY_TICK_MAX)
      {
        currtime = time(NULL);
        drvrlog = fopen(drvrlogfile,"a+");
        fprintf(drvrlog,"%.24s: Thread %d timed out\n",
                asctime(localtime(&currtime)),
                i);
        fclose(drvrlog);
        sys$cancel(t->channel);
      }
    }
  }
}

main()
{
  init();
  for (;;)
  {
    drain_queue();
    if(inactivity_tick_trap_flag)
	check_for_inactive_threads();
    (void)sys$hiber();
  }
}

/*************************************************************************
 * NNTP_SERVER interface routines                                        *
 *************************************************************************/

int read_net(string,size,thread)
char *string;
int size, thread;
{
  struct thread_struct *t = &threads[thread];
  struct line_block *b = t->read_head;
  if (b)
  {
    strcpy(string,b->line); strcat(string,"\n");
    t->read_head = b->next;
    free(b->line);
    free(b);
/*  drvrlog = fopen(drvrlogfile,"a+"); */
/*  fprintf(drvrlog,"%d >>> %s",thread,string); */
/*  fclose(drvrlog); */
    return(b->len+1);
/* +1 for the CR */
  }
  return(1);
}

int write_net(string,thread)
char *string;
int thread;
{
  struct thread_struct *t = &threads[thread];
  struct line_block *b;
  if (t->active==ACT_HEALTHY)
  {
    b = (struct line_block *) malloc(sizeof(struct line_block));
    b->next = NULL;
    b->line = strdup(string);
/*  drvrlog = fopen(drvrlogfile,"a+"); */
/*  fprintf(drvrlog,"%d <<< %s",thread,string); */
/*  fclose(drvrlog); */
    if (t->write_head)
    {
      t->write_tail->next = b;
      t->write_tail = b;
    }
    else
    {
      t->write_tail = b;
      t->write_head = b;
    }
  }
}

int next_call(thread,func,type)
int thread;
int (*func)();
int type;
{
  threads[thread].next_function = func;
  threads[thread].next_type = type;
}

void getremhost(remhost,remuser,thread)
char *remhost, *remuser;
int thread;
{
  struct thread_struct *t = &threads[thread];
  struct itlst_3 rsck_adrs;
  struct mysockaddr rmtsckadr;
  struct hostent *hostinfop,hostinfo, *gethostbyaddr();

  unsigned short iosb[4];
  int status;
  strcpy(remuser,"nntp"); strcpy(t->remuser,"nntp");
  strcpy(remhost,"unknown"); strcpy(t->remhost,"unknown");

  rsck_adrs.lgth = sizeof rmtsckadr;
  rsck_adrs.hst = &rmtsckadr;
  rsck_adrs.retlth = &retlen;
  retlen=0;
  status = sys$qiow(0,threads[thread].channel,IO$_SENSEMODE,iosb,0,0,0,0,0,
                    &rsck_adrs,0,0);
  if (!(status&1)) return;
  if ((iosb[0]&1)-1) return;
  if((hostinfop=gethostbyaddr(rmtsckadr.adrs,4,AF_INET))!=NULL)
  {
  	hostinfo = *hostinfop;
        strcpy(remhost,hostinfo.h_name);
  }
/* try twice, in case the nameserver is slow */
  else if((hostinfop=gethostbyaddr(rmtsckadr.adrs,4,AF_INET))!=NULL)
  {
  	hostinfo = *hostinfop;
        strcpy(remhost,hostinfo.h_name);
  }
  else
  {
  	sprintf(remhost,"%d.%d.%d.%d",
                rmtsckadr.adrs[0],
                rmtsckadr.adrs[1],
                rmtsckadr.adrs[2],
                rmtsckadr.adrs[3]);
  }
  strcpy(t->remhost,remhost);
}
